#!/usr/bin/perl

# Module to perform functions associated with process accounting data.

# Written 2003-12-31 by Jim Lippard.
# Modified 2012-12-21 by Jim Lippard to create stats dir if not present,
#    add timestamp option to anomaly display, and to use directory of
#    the user running this code.
# Modified 2012-12-22 by Jim Lippard to add output_file option to comparisons
#    and show_lastcomm.

# To Do:
# Format of data read from data files should be the same as the format of data
# read from stat files (i.e., use arrays for flags, ttys, users, commands), and
# should make use of the stat field names that are in the arrays.

# Some performance efficiencies are direly needed, especially in acaudit.pl.

package ProcessAccounting;

# Required packages.
use strict;
use Carp;
use Time::localtime;
use Time::ParseDate;

# Global constants.
my $SYSTEM = 0;
my $COMMAND = 1;
my $USER = 2;
my $USER_COMMAND = 3;

my $USERNAME = $ENV{'LOGNAME'} || $ENV{'USER'} || 'lippard';

my $STATS_DIR = "/home/$USERNAME/stats";
my $SYSTEM_STAT_FILE = "$STATS_DIR/system.stat";
my $COMMANDS_DIR = "$STATS_DIR/commands";
my $USERS_DIR = "$STATS_DIR/users";

my $DATA_SUFFIX = '.data';
my $STAT_SUFFIX = '.stat';

my $NL = "
";

my @COMMON_STAT_FIELDS = (
			  'nentry',
			  'maxcpu', 'mincpu', 'totcpu', 'avgcpu', 'sumsqcpu', 'sdevcpu',
			  'maxdur', 'mindur', 'totdur', 'avgdur', 'sumsqdur', 'sdevdur',
			  'todear', 'todlat',
			  'flags', 'ttys'
			  );
my @COMMAND_STAT_FIELDS = (
			   'users'
			   );
my @USER_STAT_FIELDS = (
			'commands'
			);

# Public methods.

# Parse a line of lastcomm output.
sub parseline {
    my ($line) = @_;
    my ($lastcomm,
	$command, $flags, $user, $tty, $time_info,
	$cpu, $secs, $day_name, $month, $day, $time, $duration);

    if (substr ($_, length ($_) - 1, 1) eq $NL) {
	chop;
    }

    ($command, $flags, $user, $tty, $time_info) = split (/\s+/, $_, 5);
    ($cpu, $secs, $day_name, $month, $day, $time, $duration) = split (/\s+/, $time_info, 7);
    $time = $day_name . ' ' . $month . ' ' . $day . ' ' . $time;
    $time = parsedate ($time);
    $duration =~ s/^\((.*)\)$/$1/;

    $lastcomm->{command} = $command;
    $lastcomm->{flags} = $flags;
    $lastcomm->{user} = $user;
    $lastcomm->{tty} = $tty;
    $lastcomm->{cpu} = $cpu;
    $lastcomm->{time} = $time;
    $lastcomm->{duration} = &_duration_seconds ($duration);
    $lastcomm->{original} = $_;

    return ($lastcomm);
}

# Returns max of two args.
sub _max {
    my ($arg1, $arg2) = @_;

    if ($arg1 > $arg2) {
	return $arg1;
    }
    else {
	return $arg2;
    }
}

# Returns min of two args.
sub _min {
    my ($arg1, $arg2) = @_;

    if ($arg1 < $arg2) {
	return $arg1;
    }
    else {
	return $arg2;
    }
}

# Returns union of two arrays.
sub _union {
    my ($array1, $array2) = @_;
    my (@array1, @array2, $item);

    @array1 = @{$array1};
    @array2 = @{$array2};

    foreach $item (@array2) {
	if (!grep (/^$item$/, @array1)) {
	    push (@array1, $item);
	}
    }

    return @array1;
}

# Print out lastcomm information.
sub show_lastcomm {
    my ($lastcomm, $output_fh) = @_;

    if (!$output_fh) {
	$output_fh = \*STDOUT;
    }

    print $output_fh "$lastcomm->{original}\n";
}

# Add suffix.
sub _add_suffix {
    my ($file, $suffix) = @_;

    if (substr ($file, length ($file) - length ($suffix), length ($suffix)) ne $suffix) {
	$file .= $suffix;
    }

    return $file;
}

# Return list of command data files.
sub list_command_data {
    my (@files);

    @files = &_list_data ($COMMANDS_DIR);
    return @files;
}

# Return list of user data files.
sub list_user_data {
    my (@files);

    @files = &_list_data ($USERS_DIR);
    return @files;
}

# Return list of data files.
sub _list_data {
    my ($dir) = @_;
    my (@files);

    opendir (DIR, $dir) || croak "Cannot open dir $dir. $!\n";
    @files = grep (/$DATA_SUFFIX$/, readdir (DIR));
    closedir (DIR);
    return @files;
}

# Write a line of lastcomm data to a command file.
sub write_command_data {
    my ($file, $lastcomm) = @_;

    $file = &_add_suffix ($file, $DATA_SUFFIX);

    &_create_dir ($STATS_DIR, "stats");
    &_create_dir ($COMMANDS_DIR, "commands");

    open (FILE, ">>$COMMANDS_DIR/$file") || croak "Cannot write to $file. $!\n";
    print FILE "$lastcomm->{flags}, $lastcomm->{user}, $lastcomm->{tty}, $lastcomm->{cpu}, $lastcomm->{time}, $lastcomm->{duration}\n";
    close (FILE);
}

# Write a line of lastcomm data to a user file.
sub write_user_data {
    my ($file, $lastcomm) = @_;

    $file = &_add_suffix ($file, $DATA_SUFFIX);

    &_create_dir ($STATS_DIR, "stats");
    &_create_dir ($USERS_DIR, "users");

    open (FILE, ">>$USERS_DIR/$file") || croak "Cannot write to $file. $!\n";
    print FILE "$lastcomm->{command}, $lastcomm->{flags}, $lastcomm->{tty}, $lastcomm->{cpu}, $lastcomm->{time}, $lastcomm->{duration}\n";
    close (FILE);
}

# Create directory if it doesn't already exist.
sub _create_dir {
    my ($dir, $name) = @_;
    my $PERMS = 0700;

    if (!-e $dir) {
	mkdir ($dir, $PERMS) || croak "Cannot create $name dir. $!\n";
    }
}

# Read lastcomm data from a command file.
sub read_command_data {
    my ($file) = @_;
    my ($data);

    $file = &_add_suffix ($file, $DATA_SUFFIX);

    $data = &_read_common_data ($COMMAND, "$COMMANDS_DIR/$file");
    return $data;
}

# Read lastcomm data from a user file.
sub read_user_data {
    my ($file) = @_;
    my ($data);

    $file = &_add_suffix ($file, $DATA_SUFFIX);

    $data = &_read_common_data ($USER, "$USERS_DIR/$file");
    return $data;
}

# Read lastcomm data from a user file for a particular command.
sub read_user_command_data {
    my ($file, $command) = @_;
    my ($data);

    $file = &_add_suffix ($file, $DATA_SUFFIX);

    $data = &_read_common_data ($USER_COMMAND, "$USERS_DIR/$file", $command);
}

# Read lastcomm data, common.
sub _read_common_data {
    my ($type, $file, $command_search) = @_;
    my ($data, $command, $flags, $user, $tty, $cpu, $time, $duration,
	$timeofday,
	%known_user, %known_command, %known_flags, %known_tty);

    open (FILE, "$file") || croak "Cannot read data file $file. $!\n";

    $data->{nentry} = 0;
    $data->{maxcpu} = $data->{totcpu} = $data->{sdevcpu} = 0;
    $data->{todear} = "99:99:99";
    $data->{todlat} = "00:00:00";
    $data->{maxdur} = $data->{totdur} = $data->{sdevdur} = 0;
    $data->{mincpu} = 99;
    $data->{mindur} = '362439.99'; # 99:99:99.99 in seconds
    $data->{sumsqcpu} = $data->{sumsqdur} = 0;

    if ($type == $COMMAND || $type == $USER_COMMAND) {
	@{$data->{users}} = undef;
    }
    if ($type == $USER || $type == $USER_COMMAND) {
	@{$data->{commands}} = undef;
    }
    @{$data->{flags}} = undef;
    @{$data->{ttys}} = undef;
    while (<FILE>) {
	chop;
	if ($type == $COMMAND) {
	    ($flags, $user, $tty, $cpu, $time, $duration) = split (/,\s+/);
	    $known_user{$user} = 1;
	}
	elsif (($type == $USER) || ($type == $USER_COMMAND)) {
	    ($command, $flags, $tty, $cpu, $time, $duration) = split (/,\s+/);
	    $known_command{$command} = 1;
	}
	
	# Compatibility code:  If duration is in human format, convert.
	if ($duration =~ /:/) {
	    $duration = &_duration_seconds ($duration);
	}
	
	if (($type != $USER_COMMAND) || (defined ($command_search) && $command eq $command_search)) {
	    $data->{nentry}++;
	    $known_flags{$flags} = 1;
	    $known_tty{$tty} = 1;
	    $data->{maxcpu} = $cpu if ($cpu > $data->{maxcpu});
	    $data->{mincpu} = $cpu if ($cpu < $data->{mincpu});
	    $data->{totcpu} += $cpu;
	    $data->{sumsqcpu} += $cpu ** 2;
	    $timeofday = &_time_of_day ($time);
	    $data->{todear} = $timeofday if (&_earlier_time_of_day ($timeofday, $data->{todear}));
	    $data->{todlat} = $timeofday if (&_later_time_of_day ($timeofday, $data->{todlat}));
	    $data->{maxdur} = $duration if ($duration > $data->{maxdur});
	    $data->{mindur} = $duration if ($duration < $data->{mindur});
	    $data->{totdur} += $duration;
	    $data->{sumsqdur} += $duration ** 2;
	}
    }
    close (FILE);

    $data->{avgcpu} = $data->{totcpu} / $data->{nentry};
    $data->{sdevcpu} = &_sdev1 ($data->{sumsqcpu}, $data->{avgcpu}, $data->{nentry});
#    my ($debugger) = &_sdev1 ($data->{sumsqcpu}, $data->{totcpu}, $data->{nentry});
#    print "DEBUG: sdev1=$debugger, sdev2=$data->{sdevcpu}\n";
    $data->{avgdur} = $data->{totdur} / $data->{nentry};
    $data->{sdevdur} = &_sdev1 ($data->{sumsqdur}, $data->{avgdur}, $data->{nentry});

    if ($type == $COMMAND) {
	foreach $user (keys (%known_user)) {
	    push (@{$data->{users}}, $user);
	}
    }
    elsif ($type == $USER) {
	foreach $command (keys (%known_command)) {
	    push (@{$data->{commands}}, $command);
	}
    }

    foreach $flags (keys (%known_flags)) {
	push (@{$data->{flags}}, $flags);
    }
    
    foreach $tty (keys (%known_tty)) {
	push (@{$data->{ttys}}, $tty);
    }

    return $data;
}

# Does command stat file exist?
sub exists_command_stat {
    my ($file, $use_alt) = @_;

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    $file .= '.alt' if ($use_alt);

    if (-e "$COMMANDS_DIR/$file") {
	return 1;
    }
    else {
	return 0;
    }
}

# Does user stat file exist?
sub exists_user_stat {
    my ($file, $use_alt) = @_;

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    $file .= '.alt' if ($use_alt);

    if (-e "$USERS_DIR/$file") {
	return 1;
    }
    else {
	return 0;
    }
}

# Does user command stat file exist?
sub exists_user_command_stat {
    my ($user, $file, $use_alt) = @_;

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    $file .= '.alt' if ($use_alt);

    if (-e "$USERS_DIR/$user/$file") {
	return 1;
    }
    else {
	return 0;
    }
}

# Write stat data to the system file.
sub write_system_stat {
    my ($stat) = @_;
    my ($file);

    $file = $SYSTEM_STAT_FILE;

    &_write_common_stat ($SYSTEM, $SYSTEM_STAT_FILE, $stat);
}

# Write stat data to a command file.
sub write_command_stat {
    my ($file, $stat) = @_;

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    # Create $COMMANDS_DIR if it doesn't already exist.
    if (!-e $COMMANDS_DIR) {
	mkdir ($COMMANDS_DIR, 0700) || croak "Cannot create commands dir. $!\n";
    }

    &_write_common_stat ($COMMAND, "$COMMANDS_DIR/$file", $stat);
}

# Write stat data to a user file.
sub write_user_stat {
    my ($file, $stat) = @_;

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    # Create $COMMANDS_DIR if it doesn't already exist.
    if (!-e $COMMANDS_DIR) {
	mkdir ($COMMANDS_DIR, 0700) || croak "Cannot create commands dir. $!\n";
    }

    &_write_common_stat ($USER, "$USERS_DIR/$file", $stat);
}

# Write stat data to a user command file.
sub write_user_command_stat {
    my ($user, $file, $stat) = @_;

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    # Create $USERS_DIR/$user if it doesn't already exist.
    if (!-e "$USERS_DIR/$user") {
	mkdir ("$USERS_DIR/$user", 0700) || croak "Cannot create user command dir for $user. $!\n";
    }

    &_write_common_stat ($USER_COMMAND, "$USERS_DIR/$user/$file", $stat);
}

# Write common data to a stat file.
sub _write_common_stat {
    my ($type, $file, $stat) = @_;
    my ($flag, $user, $command, $tty);

    open (FILE, ">$file") || croak "Cannot write to stat file $file. $!\n";
    print FILE "nentry: $stat->{nentry}\n";
    print FILE "maxcpu: $stat->{maxcpu}\n";
    print FILE "mincpu: $stat->{mincpu}\n";
    print FILE "totcpu: $stat->{totcpu}\n";
    print FILE "avgcpu: $stat->{avgcpu}\n";
    print FILE "sumsqcpu: $stat->{sumsqcpu}\n";
    print FILE "sdevcpu: $stat->{sdevcpu}\n";
    print FILE "todear: $stat->{todear}\n";
    print FILE "todlat: $stat->{todlat}\n";
    print FILE "maxdur: $stat->{maxdur}\n";
    print FILE "mindur: $stat->{mindur}\n";
    print FILE "totdur: $stat->{totdur}\n";
    print FILE "avgdur: $stat->{avgdur}\n";
    print FILE "sumsqdur: $stat->{sumsqdur}\n";
    print FILE "sdevdur: $stat->{sdevdur}\n";
    print FILE "flags:";
    foreach $flag (@{$stat->{flags}}) {
	print FILE " $flag" if ($flag ne "");
    }
    print FILE "\n";
    if ($type == $SYSTEM || $type == $COMMAND) {
	print FILE "users:";
	foreach $user (@{$stat->{users}}) {
	    print FILE " $user" if ($user ne "");
	}
	print FILE "\n";
    }
    if ($type == $SYSTEM || $type == $USER) {
	print FILE "commands:";
	foreach $command (@{$stat->{commands}}) {
	    print FILE " $command" if ($command ne "");
	}
	print FILE "\n";
    }
    print FILE "ttys:";
    foreach $tty (@{$stat->{ttys}}) {
	print FILE " $tty" if ($tty ne "");
    }
    print FILE "\n";
    close (FILE);
}

# Merge two system stats together.
sub merge_system_stats {
    my ($stat1, $stat2) = @_;
    my ($return_stat);

    $return_stat = &_merge_common_stats ($SYSTEM, $stat1, $stat2);
    return $return_stat;
}

# Merge two command stats together.
sub merge_command_stats {
    my ($stat1, $stat2) = @_;
    my ($return_stat);

    $return_stat = &_merge_common_stats ($COMMAND, $stat1, $stat2);
    return $return_stat;
}

# Merge two user stats together.
sub merge_user_stats {
    my ($stat1, $stat2) = @_;
    my ($return_stat);

    $return_stat = &_merge_common_stats ($USER, $stat1, $stat2);
    return $return_stat;
}

# Merge two user command stats together.
sub merge_user_command_stats {
    my ($stat1, $stat2) = @_;
    my ($return_stat);

    $return_stat = &_merge_common_stats ($USER_COMMAND, $stat1, $stat2);
}

# Common code for merging stats.
# Merging means taking the unions of lists, the maxes and mins in the comparisons,
# adding the totals and sums of squares, calculating new averages and standard deviations.
sub _merge_common_stats {
    my ($type, $stat1, $stat2) = @_;
    my ($return_stat);

    $return_stat->{nentry} = $stat1->{nentry} + $stat2->{nentry};

    # CPU.
    $return_stat->{maxcpu} = &_max ($stat1->{maxcpu}, $stat2->{maxcpu});
    $return_stat->{mincpu} = &_min ($stat1->{mincpu}, $stat2->{mincpu});
    $return_stat->{totcpu} = $stat1->{totcpu} + $stat2->{totcpu};
    $return_stat->{avgcpu} = $return_stat->{totcpu} / $return_stat->{nentry};
    $return_stat->{sumsqcpu} = $stat1->{sumsqcpu} + $stat2->{sumsqcpu};
    $return_stat->{sdevcpu} = &_sdev1 ($return_stat->{sumsqcpu}, $return_stat->{avgcpu}, $return_stat->{nentry});

    # Time of day.
    if (&_earlier_time_of_day ($stat1->{todear}, $stat2->{todear})) {
	$return_stat->{todear} = $stat1->{todear};
    }
    else {
	$return_stat->{todear} = $stat2->{todear};
    }
    if (&_later_time_of_day ($stat1->{todlat}, $stat2->{todlat})) {
	$return_stat->{todlat} = $stat1->{todlat};
    }
    else {
	$return_stat->{todlat} = $stat2->{todlat};
    }

    # Durations.
    $return_stat->{maxdur} = &_max ($stat1->{maxdur}, $stat2->{maxdur});
    $return_stat->{mindur} = &_min ($stat1->{mindur}, $stat2->{mindur});
    $return_stat->{totdur} = $stat1->{totdur} + $stat2->{totdur};
    $return_stat->{avgdur} = $return_stat->{totdur} / $return_stat->{nentry};
    $return_stat->{sumsqdur} = $stat1->{sumsqcpu} + $stat2->{sumsqcpu};
    $return_stat->{sdevdur} = &_sdev1 ($return_stat->{sumsqdur}, $return_stat->{avgdur}, $return_stat->{nentry});

    # Flags.
    @{$return_stat->{flags}} = &_union ($stat1->{flags}, $stat2->{flags});

    # Users.
    if ($type == $SYSTEM || $type == $COMMAND) {
	@{$return_stat->{users}} = &_union ($stat1->{users}, $stat2->{users});
    }
    # Commands.
    elsif ($type == $SYSTEM || $type == $USER) {
	@{$return_stat->{commands}} = &_union ($stat1->{commands}, $stat2->{commands});
    }

    # TTYs.
    @{$return_stat->{ttys}} = &_union ($stat1->{ttys}, $stat2->{ttys});

    return $return_stat;
}

# Read system stat from the system stat file.
sub read_system_stat {
    my ($use_alt) = @_;
    my ($file, $stat);

    $file = $SYSTEM_STAT_FILE;

    if ($use_alt && &exists_command_stat ($file, $use_alt)) {
	$file .= '.alt';
    }

    $stat = &_read_common_stat ($SYSTEM, $SYSTEM_STAT_FILE);
    return $stat;
}

# Read stat data from a command file.
sub read_command_stat {
    my ($file, $use_alt) = @_;
    my ($stat);

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    if ($use_alt && &exists_command_stat ($file, $use_alt)) {
	$file .= '.alt';
    }

    $stat = &_read_common_stat ($COMMAND, "$COMMANDS_DIR/$file");
    return $stat;
}

# Read stat data from a user file.
sub read_user_stat {
    my ($file, $use_alt) = @_;
    my ($stat);

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    if ($use_alt && &exists_user_stat ($file, $use_alt)) {
	$file .= '.alt';
    }

    $stat = &_read_common_stat ($USER, "$USERS_DIR/$file");
    return $stat;
}

# Read stat data from a user command file.
sub read_user_command_stat {
    my ($user, $file, $use_alt) = @_;
    my ($stat);

    $file = &_add_suffix ($file, $STAT_SUFFIX);

    if ($use_alt && &exists_user_command_stat ($user, $file, $use_alt)) {
	$file .= '.alt';
    }

    $stat = &_read_common_stat ($USER_COMMAND, "$USERS_DIR/$user/$file");
    return $stat;
}

# Common stat reading.  User command stats have only the common stat fields.
sub _read_common_stat {
    my ($type, $file) = @_;
    my ($stat, $field, $value);

    open (FILE, "$file") || croak "Cannot read stat file $file. $!\n";
    while (<FILE>) {
	($field, $value) = split (/\:\s+/);
	chop ($value);
	if (grep (/^$field$/, @COMMON_STAT_FIELDS) ||
	    (($type == $SYSTEM || $type == $COMMAND) && grep (/^$field$/, @COMMAND_STAT_FIELDS)) ||
	    (($type == $SYSTEM || $type == $USER) && grep (/^$field$/, @USER_STAT_FIELDS))) {
	    if ($field eq 'flags' || $field eq 'ttys' || $field eq 'commands' || $field eq 'users') {
		@{$stat->{$field}} = split (/\s+/, $value);
	    }
	    else {
		$stat->{$field} = $value;
	    }
	}
	else {
	    croak "Unidentified field $field in stat file $file.\n";
	}
    }
    close (FILE);

    return $stat;
}

# Ignore matches when a field is set to "*".  Only look at the first entry.
sub _ignore_matches {
    my (@args) = @_;
    my ($arg);

    $arg = shift (@args);
    if ($arg eq '*') {
	return 1;
    }
    else {
	return 0;
    }
}

# Compare lastcomm data to system stat data.  (Either generate output or return discrepant items.)
sub compare_to_system_stat {
    my ($lastcomm, $stat, $quiet, $timestamp, $output_fh) = @_;
    my ($anomalous);

    $anomalous = &_compare_to_stat ($SYSTEM, $lastcomm, $stat, $quiet, $timestamp, $output_fh);
    return $anomalous;
}

# Compare lastcomm data to command stat data.  (Either generate output or return discrepant items.)
sub compare_to_user_stat {
    my ($lastcomm, $stat, $quiet, $timestamp, $output_fh) = @_;
    my ($anomalous);

    $anomalous = &_compare_to_stat ($USER, $lastcomm, $stat, $quiet, $timestamp, $output_fh);
    return $anomalous;
}

# Compare lastcomm data to user stat data.  (Either generate output or return discrepant items.)
sub compare_to_command_stat {
    my ($lastcomm, $stat, $quiet, $timestamp, $output_fh) = @_;
    my ($anomalous);

    $anomalous = &_compare_to_stat ($COMMAND, $lastcomm, $stat, $quiet, $timestamp, $output_fh);
    return $anomalous;
}

# Compare lastcomm data to user command stat data.  (Either generate output or return discrepant items.)
sub compare_to_user_command_stat {
    my ($lastcomm, $stat, $quiet, $timestamp, $output_fh) = @_;
    my ($anomalous);

    $anomalous = &_compare_to_stat ($USER_COMMAND, $lastcomm, $stat, $quiet, $timestamp, $output_fh);
    return $anomalous;
}

# Common compare.
sub _compare_to_stat {
    my ($type, $lastcomm, $stat, $quiet, $timestamp, $output_fh) = @_;
    my ($anomalous, $timeofday, $duration_human);

    if ($type == $SYSTEM || $type == $USER) {
	if (!&_ignore_matches (@{$stat->{commands}}) && !grep (/^$lastcomm->{command}$/, @{$stat->{commands}})) {
	    $anomalous = 1;
	    return $anomalous if ($quiet);
	    # We lie about type to show_anomaly to get the right grammar here.
	    if ($type == $SYSTEM) {
		&show_anomaly ($COMMAND, $lastcomm, " by anyone", $timestamp, $output_fh);
	    }
	    elsif ($type == $USER) {
		&show_anomaly ($COMMAND, $lastcomm, " by user $lastcomm->{user}", $timestamp, $output_fh);
	    }
	}
    }

    if ($type == $SYSTEM) {
	if (!&_ignore_matches (@{$stat->{users}}) && !grep (/^$lastcomm->{user}$/, @{$stat->{users}})) {
	    $anomalous = 1;
	    return $anomalous if ($quiet);
	    # We fake the type because we want the form "User {user} has not previously executed any commands."
	    &show_anomaly ($USER, $lastcomm, "", $timestamp, $output_fh);
	}
    }
    elsif ($type == $COMMAND) {
	if (!&_ignore_matches (@{$stat->{users}}) && !grep (/^$lastcomm->{user}$/, @{$stat->{users}})) {
	    $anomalous = 1;
	    return $anomalous if ($quiet);
	    &show_anomaly ($type, $lastcomm, " by user $lastcomm->{user}", $timestamp, $output_fh);
	}
    }
    # No need to look at user or command for user command stat.

    if (!&_ignore_matches (@{$stat->{flags}}) && !grep (/^$lastcomm->{flags}$/, @{$stat->{flags}})) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	&show_anomaly ($type, $lastcomm, " with flags $lastcomm->{flags}", $timestamp, $output_fh);
    }

    if (!&_ignore_matches (@{$stat->{ttys}}) && !grep (/^$lastcomm->{tty}$/, @{$stat->{ttys}})) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	&show_anomaly ($type, $lastcomm, " from tty $lastcomm->{tty}", $timestamp, $output_fh);
    }

    if (!&_ignore_matches ($stat->{maxcpu}) && ($lastcomm->{cpu} > ($stat->{maxcpu} + $stat->{sdevcpu}))) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	&show_anomaly ($type, $lastcomm, " which took more than $stat->{maxcpu} cpu seconds", $timestamp, $output_fh);
    }
    elsif (!&_ignore_matches ($stat->{mincpu}) && ($lastcomm->{cpu} < ($stat->{mincpu} - $stat->{sdevcpu}))) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	&show_anomaly ($type, $lastcomm, " which took less than $stat->{mincpu} cpu seconds", $timestamp, $output_fh);
    }

    # Time comparisons need some additional thought and coding.
    $timeofday = &_time_of_day ($lastcomm->{time});
    if (!&_ignore_matches ($stat->{todear}) && &_earlier_time_of_day ($timeofday, $stat->{todear})) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	&show_anomaly ($type, $lastcomm, " earlier than $stat->{todear}", $timestamp, $output_fh);
    }

    if (!&_ignore_matches ($stat->{todlat}) && &_later_time_of_day ($timeofday, $stat->{todlat})) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	&show_anomaly ($type, $lastcomm, " later than $stat->{todlat}", $timestamp, $output_fh);
    }

    if (!&_ignore_matches ($stat->{maxdur}) && ($lastcomm->{duration} > ($stat->{maxdur} + $stat->{sdevdur}))) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	$duration_human = &_duration_human ($stat->{maxdur});
	&show_anomaly ($type, $lastcomm, " which took more than $duration_human duration", $timestamp, $output_fh);
    }
    elsif (!&_ignore_matches ($stat->{mindur}) && ($lastcomm->{duration} < ($stat->{mindur} - $stat->{sdevdur}))) {
	$anomalous = 1;
	return $anomalous if ($quiet);
	$duration_human = &_duration_human ($stat->{mindur});
	&show_anomaly ($type, $lastcomm, " which took less than $duration_human duration", $timestamp, $output_fh);
    }

    return $anomalous;
}

# Subroutine to display anomaly information in appropriate format.
sub show_anomaly {
    my ($type, $lastcomm, $message, $timestamp, $output_fh) = @_;
    my ($tm_obj);

    if (!$output_fh) {
	$output_fh = \*STDOUT;
    }

    if ($timestamp) {
	# ctime() is just a way to get to the default localtime() function
	# which has been overridden by Time::localtime.
#	$time = ctime (time());
#	print "$time ";
	$tm_obj = localtime (time());
	printf $output_fh "%d-%02d-%02d %02d:%02d:%02d ", $tm_obj->year() + 1900, $tm_obj->mon() + 1, $tm_obj->mday(), $tm_obj->hour(), $tm_obj->min(), $tm_obj->sec(); 
    }

    if ($type == $SYSTEM) {
	print $output_fh "No one has previously executed any commands$message.\n";
    }
    elsif ($type == $USER) {
	print $output_fh "User $lastcomm->{user} has not previously executed any commands$message.\n";
    }
    elsif ($type == $COMMAND) {
	print $output_fh "Command $lastcomm->{command} has not previously been executed$message.\n";
    }
    elsif ($type == $USER_COMMAND) {
	print $output_fh "User $lastcomm->{user} has not previously executed command $lastcomm->{command}$message.\n";
    }
}

### Subroutine to calculate standard deviation using aggregate data only.

# This version uses the sum of the squares and the sum.
sub _sdev1 {
    my ($sum_of_squares, $sum, $n_entries) = @_;
    my ($variance, $sdev);

    if ($n_entries < 2) {
	return 0;
    }
    $variance = ($sum_of_squares - (($sum ** 2) / $n_entries)) / ($n_entries - 1);
    if ($variance < 0) {
	return 1;
    }
    $sdev = sqrt ($variance);
    return $sdev;
}

# This version uses the sum of the squares and the average.
sub _sdev2 {
    my ($sum_of_squares, $average, $n_entries) = @_;
    my ($variance, $sdev);

    $variance = ($sum_of_squares / $n_entries) - ($average ** 2);
    if ($variance < 0) {
	return 1;
    }
    $sdev = sqrt ($variance);

    return $sdev;
}

### Subroutines to deal with times of day.

# Convert time to time of day.
sub _time_of_day {
    my ($time) = @_;
    my ($tm_obj, $hour, $min, $sec);

    $tm_obj = localtime ($time);
    $hour = sprintf ("%02d", $tm_obj->hour());
#    $hour = $tm_obj->hour();
#    if (length ($hour) == 1) {
#	$hour = "0$hour";
#    }
    $min = sprintf ("%02d", $tm_obj->min());
#    $min = $tm_obj->min();
#    if (length ($min) == 1) {
#	$min = "0$min";
#    }
    $sec = sprintf ("%02d", $tm_obj->sec());
#    $sec = $tm_obj->sec();
#    if (length ($sec) == 1) {
#	$sec = "0$sec";
#    }
    return "$hour:$min:$sec";
}

# Return 1 if $time1 < $time2 (looking only at time of day).
sub _earlier_time_of_day {
    my ($time1, $time2) = @_;
    my ($hour1, $min1, $sec1, $hour2, $min2, $sec2);

    ($hour1, $min1, $sec1) = split (/:/, $time1);
    ($hour2, $min2, $sec2) = split (/:/, $time2);

    if ($hour1 < $hour2) {
	return 1;
    }
    elsif ($hour1 > $hour2) {
	return 0;
    }
    elsif ($min1 < $min2) {
	return 1;
    }
    elsif ($min1 > $min2) {
	return 0;
    }
    elsif ($sec1 < $sec2) {
	return 1;
    }

    return 0;
}

# Return 1 if $time1 > $time2 (looking only at time of day).
sub _later_time_of_day {
    my ($time1, $time2) = @_;

    return (&_earlier_time_of_day ($time2, $time1));
}

### Subroutines to deal with durations.

# Subroutine to convert duration to seconds.
sub _duration_seconds {
    my ($duration) = @_;
    my ($hours, $minutes, $seconds, $result_seconds);

    ($hours, $minutes, $seconds) = split (/:/, $duration, 3);
    $result_seconds = $seconds + $minutes * 60 + $hours * 60 * 60;
    return $result_seconds;
}

# Subroutine to convert duration from seconds to human format.
sub _duration_human {
    my ($duration) = @_;
    my ($hours, $minutes, $seconds, $mantissa, $return_duration);

    $hours = $minutes = $seconds = 0;

    if ($duration > 60) {
	$minutes = int ($duration / 60);
	$seconds = $duration % 60;
    }
    $seconds = sprintf ("%02d", $seconds);
#    if (length ($seconds) == 1) {
#	$seconds = "0$seconds";
#    }

    if ($minutes > 60) {
	$hours = int ($minutes / 60);
	$minutes = $minutes % 60;
    }
    $minutes = sprintf ("%02d", $minutes);
#    if (length ($minutes) == 1) {
#	$minutes = "0$minutes";
#    }

    $mantissa = $duration - int ($duration);
    $mantissa =~ s/^0\.//;

    $return_duration = "$hours:$minutes:$seconds.$mantissa";
    return $return_duration;
}

1;
