#!/usr/bin/perl

# Written 2003-12-31 by Jim Lippard.
# Modified 2012-12-21 by Jim Lippard to add -v and -t options.
# Modified 2012-12-22 by Jim Lippard to add -f option.

use strict;
use Getopt::Std;
use ProcessAccounting;
use Time::ParseDate;

# Tool to report discrepancies from saved process accounting information.

# Options:
# -q quiet, do not show output reporting anomalies, just anomalous lastcomm lines.
# -v verbose, show timestamp on output reporting anomalies.
# -t starttime, stop when you hit lastcomm output at or earlier than starttime.

# Each line of output from lastcomm looks like this:

#Command    Flags   User                             TTY        CPU       Time             Duration
#calendar   -       lippard                          ttyp2      0.00 secs Wed Dec 31 12:53 (0:00:00.02)
#cpp        -       lippard                          ttyp2      0.00 secs Wed Dec 31 12:53 (0:00:00.02)
#cpp0       -       lippard                          ttyp2      0.00 secs Wed Dec 31 12:53 (0:00:00.00)
#stty       -       lippard                          ttyp2      0.00 secs Wed Dec 31 12:53 (0:00:00.00)
#bash       -F      lippard                          ttyp2      0.00 secs Wed Dec 31 12:53 (0:00:00.00)
#sed        -       lippard                          ttyp2      0.00 secs Wed Dec 31 12:53 (0:00:00.00)
#bash       -F      lippard                          ttyp2      0.00 secs Wed Dec 31 12:53 (0:00:00.00)

# This tool will read lines of lastcomm output from STDIN, and for each line,
# either generate no output or generate output to note discrepancies:
#
# 1.  Is this a command that this user has not previously executed?
# 2.  Is this command being executed with flags it has not previously been executed with by anyone?
# 3.  Is this command being executed with flags it has not previously been executed with by this user?
# 4.  Is this command being executed from a TTY from which it has not previously been executed?
# 5.  Is this user on a TTY he has not previously executed commands from?
# 6.  Is this command using an abnormally high or low amount of CPU time?
# 7.  Is this command using an abnormally high or low amount of CPU time for this user?
# 8.  Is this command being executed at a time of day outside of previous ranges for this command?  (tough one)
# 9.  Is this command being executed at a time of day outside of previous ranges for this user?  (tough one)
# 10.  Did this command's execution last for an abnormally long or short amount of time for this command?
# 11.  Did this command's execution last for an abnormally long or short amount of time for this user?

# Given the current collection of data, I cannot answer the per-user questions.

my $USE_ALT = 1;
my $DONT_USE_ALT = 0;

my (%options, $quiet, $verbose, $start_time, $output_file, $output_fh, $lastcomm, $file, $system_stat, $command_stat, $user, $user_stat, $user_command_stat, $anomalous, $returned_anomalous);

# Get options.
getopts ('qt:vf:', \%options);

if ($#ARGV >= 0) {
    die "Usage: acaudit.pl [-q|-t starttime|-v|-f output-file]\n";
}

$quiet, $verbose, $start_time = 0;

if ($options{'q'}) {
    $quiet = 1;
}

if ($options{'v'}) {
    $verbose = 1;
}

if ($options{'t'}) {
    $start_time = $options{'t'};
    $start_time = parsedate ($start_time);
}

if ($options{'f'}) {
    $output_file = $options{'f'};
}

$anomalous, $returned_anomalous = 0;

if ($output_file) {
    open (OUTPUT_FILE, ">>$output_file") || die "Cannot open $output_file for writing. $!\n";
    $output_fh = \*OUTPUT_FILE;
}
else {
    $output_fh = \*STDOUT;
}

while (<STDIN>) {
    $lastcomm = ProcessAccounting::parseline ($_);

    # If time of this entry is equal to or earlier than start time, we're done.
    next if ($lastcomm->{time} <= $start_time); # last causes STDIN pipe issues and aborts.

    # Check system stat.
    $system_stat = &ProcessAccounting::read_system_stat ($USE_ALT);
    $anomalous = &ProcessAccounting::compare_to_system_stat ($lastcomm, $system_stat, $quiet, $verbose, $output_fh);

    # No need to make any further comparisons if $quiet.
    if ($anomalous && $quiet) {
	&ProcessAccounting::show_lastcomm ($lastcomm, $output_fh);
	next;
    }

    # Check command stat.
    $file = "$lastcomm->{command}.stat";
    if (&ProcessAccounting::exists_command_stat ($file, $DONT_USE_ALT)) {
	$command_stat = &ProcessAccounting::read_command_stat ($file, $USE_ALT);
	$returned_anomalous = &ProcessAccounting::compare_to_command_stat ($lastcomm, $command_stat, $quiet, $verbose, $output_fh);
	$anomalous = $anomalous || $returned_anomalous;
    }

    # No need to make any further comparisons if $quiet.
    if ($anomalous && $quiet) {
	&ProcessAccounting::show_lastcomm ($lastcomm, $output_fh);
	next;
    }

    # Check user stat.
    $file = "$lastcomm->{user}.stat";
    if (&ProcessAccounting::exists_user_stat ($file, $DONT_USE_ALT)) {
	$user_stat = &ProcessAccounting::read_user_stat ("$lastcomm->{user}.stat", $USE_ALT);
	$returned_anomalous = &ProcessAccounting::compare_to_user_stat ($lastcomm, $user_stat, $quiet, $verbose, $output_fh);
	$anomalous = $anomalous || $returned_anomalous;
    }

    # No need to make any further comparisons if $quiet.
    if ($anomalous && $quiet) {
	&ProcessAccounting::show_lastcomm ($lastcomm, $output_fh);
	next;
    }

    # Check user command stat.
    $file = "$lastcomm->{command}.stat";
    if (!&ProcessAccounting::exists_user_command_stat ($lastcomm->{user}, $file, $DONT_USE_ALT)) {
# This is redundant.
#	print "User $lastcomm->{user} has not previously executed command $lastcomm->{command}.\n" if (!$quiet);
#	$anomalous = 1;
    }
    else {
	$user_command_stat = &ProcessAccounting::read_user_command_stat ($lastcomm->{user}, "$lastcomm->{command}.stat", $USE_ALT);
	$returned_anomalous = &ProcessAccounting::compare_to_user_command_stat ($lastcomm, $user_command_stat, $quiet, $verbose, $output_fh);
	$anomalous = $anomalous || $returned_anomalous;
    }

    if ($anomalous) {
	&ProcessAccounting::show_lastcomm ($lastcomm, $output_fh);
    }
}

if ($output_file) {
    close (OUTPUT_FILE);
}
