#!/usr/bin/perl

# Script to monitor process accounting, report deviances from
# baseline, and update the baseline (when in tuning mode), or
# leaves the baseline alone (when in production mode).

# Written 2012-12-21 by Jim Lippard.

### Requirements.
use strict;
use Getopt::Std;

### Constants.
my $ACCOLLECT = '/usr/local/bin/accollect.pl';
my $ACSTATS = '/usr/local/bin/acstats.pl';
my $ACAUDIT = '/usr/local/bin/acaudit.pl';
my $LASTCOMM = '/usr/bin/lastcomm';
my $MKTEMP = '/usr/bin/mktemp';

my $USERNAME = $ENV{'LOGNAME'} || $ENV{'USER'} || 'lippard';

my $LAST_RUNTIME_FILE = "/home/$USERNAME/stats/acct_monitor.last";
my $LOGFILE = "/home/$USERNAME/stats/acct_monitor.log";

### Variables.
my (%options, $production_mode, $last_runtime, $new_last_runtime, $merge_flag, $deleted_data, $tmpfile);

### Program.
$deleted_data = 0;
$production_mode = 0;

getopts ('p', \%options);

if ($options{'p'}) {
    $production_mode = 1;
}

if (open (FILE, "$LAST_RUNTIME_FILE")) {
    $last_runtime = <FILE>;
    chop ($last_runtime);
    if ($last_runtime =~ /:/) {
	($last_runtime, $merge_flag) = split (/:/, $last_runtime);
    }
    $last_runtime = localtime ($last_runtime);
}
else {
    $last_runtime = 0;
}

$tmpfile = `$MKTEMP -q /tmp/acct_monitor.XXXXXXX`;
chop ($tmpfile);

# We open up a temp file, and write the lastcomm output to it, which
# we also feed to accollect.pl, stopping at the last_runtime if nonzero.
# We then do an acstats.pl to update the baseline, and then feed the
# same lastcomm output again to acaudit.pl, writing the output to a log
# file.
#
# We could, if we wanted, send the lastcomm output twice to acaudit.pl,
# once to collect the raw lastcomm output with divergences from the
# current baseline (with -q, also with -t start_time), and again to
# display the output, but that would make the averages wrong.
#
# But periodically, the old data files should be deleted and then new
# data merged via actstat.pl -m on the first subsequent run.  Perhaps
# based on data file size.  That could be done with this program, saving
# a flag in $LAST_RUNTIME_FILE so it knows what to do on the next run.

## HERE: When in tuning mode (and data files grow), if they are really big, delete and set $data_deleted = 1.

# Open the temp file to save raw lastcomm output.
open (TEMP, ">$tmpfile") || die "Cannot write to temp file $tmpfile. $!\n";

# Open lastcomm.
open (LASTCOMM, "$LASTCOMM |") || die "Cannot open $LASTCOMM. $!\n";
if ($last_runtime && !$production_mode) {
    open (ACCTCOLLECT, "| $ACCOLLECT -t \"$last_runtime\"") || die "Cannot open $ACCOLLECT. $!\n";
}
elsif (!$production_mode) {
    open (ACCTCOLLECT, "| $ACCOLLECT") || die "Cannot open $ACCOLLECT. $!\n";
}

while (<LASTCOMM>) {
    print TEMP $_;
    print ACCTCOLLECT $_ if (!$production_mode);  # This aborts if -t is used and implemented in a way that prematurely closes the pipe. Fixed in accollect.pl.
}
close (ACCTCOLLECT) if (!$production_mode);
close (LASTCOMM);
close (TEMP);

# Save current time as last runtime.  There's some possibility to
# miss some lastcomm data here.
open (FILE, ">$LAST_RUNTIME_FILE") || die "Could not open file $LAST_RUNTIME_FILE for writing. $!\n";
$new_last_runtime = time();
print FILE "$new_last_runtime";
if ($deleted_data && !$production_mode) { # set merge_flag for next run
    print FILE ":MERGE";
}
print FILE "\n";
close (FILE);

# Open the tempfile to read back out the lastcomm data, send it to acaudit.pl
# with the output directed to the logfile.
open (TEMP, $tmpfile) || die "Cannot open temp file $tmpfile for reading. $!\n";

if ($last_runtime) {
    open (ACCTAUDIT, "| $ACAUDIT -v -t \"$last_runtime\" -f $LOGFILE") || die "Cannot open $ACAUDIT. $!\n";
}
else {
    open (ACCTAUDIT, "| $ACAUDIT -v -f $LOGFILE") || die "Cannot open $ACAUDIT. $!\n";
}

# Send contents of temp file to ACAUDIT.
while (<TEMP>) {
    print ACCTAUDIT $_;
}
close (ACCTAUDIT);
close (TEMP);

# Clean up temp file.
unlink ($tmpfile);

# Now update the baseline with acstats.pl.  Must do this after acaudit.pl
# or the stats baseline will already reflect the new data.
if ($merge_flag) {
    system "$ACSTATS -m";
}
else {
    system "$ACSTATS";
}

